# https://nomads.ncep.noaa.gov/gribfilter.php?ds=gfs_0p25_1hr
gfs_parameters_dict = {
    'var_4LFTX': 'off',
    'var_ABSV': 'off',
    'var_ACPCP': 'off',
    'var_ALBDO': 'off',
    'var_APCP': 'off',
    'var_CAPE': 'off',
    'var_CFRZR': 'off',
    'var_CICEP': 'off',
    'var_CIN': 'off',
    'var_CLWMR': 'off',
    'var_CNWAT': 'off',
    'var_CPOFP': 'off',
    'var_CPRAT': 'off',
    'var_CRAIN': 'off',
    'var_CSNOW': 'off',
    'var_CWAT': 'off',
    'var_CWORK': 'off',
    'var_DLWRF': 'off',
    'var_DPT': 'off',
    'var_DSWRF': 'off',
    'var_DZDT': 'off',
    'var_FLDCP': 'off',
    'var_FRICV': 'off',
    'var_GFLUX': 'off',
    'var_GRLEV': 'off',
    'var_GUST': 'off',
    'var_HCDC': 'off',
    'var_HGT': 'off',
    'var_HINDEX': 'off',
    'var_HLCY': 'off',
    'var_HPBL': 'off',
    'var_ICAHT': 'off',
    'var_ICEC': 'off',
    'var_ICEG': 'off',
    'var_ICETK': 'off',
    'var_ICETMP': 'off',
    'var_ICMR': 'off',
    'var_LAND': 'off',
    'var_LCDC': 'off',
    'var_LFTX': 'off',
    'var_LHTFL': 'off',
    'var_MCDC': 'off',
    'var_MSLET': 'off',
    'var_O3MR': 'off',
    'var_PEVFR': 'off',
    'var_PLPL': 'off',
    'var_POT': 'off',
    'var_PRATE': 'off',
    'var_PRES': 'off',
    'var_PRMSL': 'off',
    'var_PWAT': 'off',
    'var_REFC': 'off',
    'var_REFD': 'off',
    'var_RH': 'off',
    'var_RWMR': 'off',
    'var_SFCR': 'off',
    'var_SHTFL': 'off',
    'var_SNMR': 'off',
    'var_SNOD': 'off',
    'var_SOILL': 'off',
    'var_SOILW': 'off',
    'var_SOTYP': 'off',
    'var_SPFH': 'off',
    'var_SUNSD': 'off',
    'var_TCDC': 'off',
    'var_TMAX': 'off',
    'var_TMIN': 'off',
    'var_TMP': 'off',
    'var_TOZNE': 'off',
    'var_TSOIL': 'off',
    'var_UFLX': 'off',
    'var_UGRD': 'off',
    'var_U-GWD': 'off',
    'var_ULWRF': 'off',
    'var_USTM': 'off',
    'var_USWRF': 'off',
    'var_VEG': 'off',
    'var_VFLX': 'off',
    'var_VGRD': 'off',
    'var_V-GWD': 'off',
    'var_VIS': 'off',
    'var_VRATE': 'off',
    'var_VSTM': 'off',
    'var_VVEL': 'off',
    'var_VWSH': 'off',
    'var_WATR': 'off',
    'var_WEASD': 'off',
    'var_WILT': 'off',
}

gfs_levels_dict = {
    'lev_0-0.1_m_below_ground': 'off',
    'lev_0.1-0.4_m_below_ground': 'off',
    'lev_0.4-1_m_below_ground': 'off',
    'lev_1-2_m_below_ground': 'off',
    'lev_2_m_above_ground': 'off',
    'lev_10_m_above_ground': 'off',
    'lev_20_m_above_ground': 'off',
    'lev_30_m_above_ground': 'off',
    'lev_40_m_above_ground': 'off',
    'lev_50_m_above_ground': 'off',
    'lev_80_m_above_ground': 'off',
    'lev_100_m_above_ground': 'off',
    'lev_1000_m_above_ground': 'off',
    'lev_4000_m_above_ground': 'off',
    'lev_10_m_above_mean_sea_level': 'off',
    'lev_1829_m_above_mean_sea_level': 'off',
    'lev_2743_m_above_mean_sea_level': 'off',
    'lev_3658_m_above_mean_sea_level': 'off',
    'lev_3000-0_m_above_ground': 'off',
    'lev_6000-0_m_above_ground': 'off',
    'lev_180-0_mb_above_ground': 'off',
    'lev_255-0_mb_above_ground': 'off',
    'lev_90-0_mb_above_ground': 'off',
    'lev_30-0_mb_above_ground': 'off',
    'lev_0C_isotherm': 'off',
    'lev_1_hybrid_level': 'off',
    'lev_2_hybrid_level': 'off',
    'lev_1000_mb': 'off',
    'lev_975_mb': 'off',
    'lev_950_mb': 'off',
    'lev_925_mb': 'off',
    'lev_900_mb': 'off',
    'lev_850_mb': 'off',
    'lev_800_mb': 'off',
    'lev_750_mb': 'off',
    'lev_700_mb': 'off',
    'lev_650_mb': 'off',
    'lev_600_mb': 'off',
    'lev_550_mb': 'off',
    'lev_500_mb': 'off',
    'lev_450_mb': 'off',
    'lev_400_mb': 'off',
    'lev_350_mb': 'off',
    'lev_300_mb': 'off',
    'lev_250_mb': 'off',
    'lev_200_mb': 'off',
    'lev_150_mb': 'off',
    'lev_100_mb': 'off',
    'lev_70_mb': 'off',
    'lev_50_mb': 'off',
    'lev_40_mb': 'off',
    'lev_30_mb': 'off',
    'lev_20_mb': 'off',
    'lev_15_mb': 'off',
    'lev_10_mb': 'off',
    'lev_7_mb': 'off',
    'lev_5_mb': 'off',
    'lev_3_mb': 'off',
    'lev_2_mb': 'off',
    'lev_1_mb': 'off',
    'lev_0.7_mb': 'off',
    'lev_0.4_mb': 'off',
    'lev_0.2_mb': 'off',
    'lev_0.1_mb': 'off',
    'lev_0.07_mb': 'off',
    'lev_0.04_mb': 'off',
    'lev_0.02_mb': 'off',
    'lev_0.01_mb': 'off',
    'lev_surface': 'off',
    'lev_mean_sea_level': 'off',
    'lev_boundary_layer_cloud_layer': 'off',
    'lev_planetary_boundary_layer': 'off',
    'lev_tropopause': 'off',
    'lev_entire_atmosphere': 'off',
    'lev_entire_atmosphere_(considered_as_a_single_layer)': 'off',
    'lev_top_atmosphere': 'off',
    'lev_convective_cloud_layer': 'off',
    'lev_convective_bottom_level': 'off',
    'lev_convective_top_level': 'off',
    'lev_high_cloud_layer': 'off',
    'lev_high_cloud_bottom_level': 'off',
    'lev_high_cloud_top_level': 'off',
    'lev_low_cloud_layer': 'off',
    'lev_low_cloud_bottom_level': 'off',
    'lev_low_cloud_top_level': 'off',
    'lev_middle_cloud_layer': 'off',
    'lev_middle_cloud_bottom_level': 'off',
    'lev_middle_cloud_top_level': 'off',
    'lev_cloud_ceiling': 'off',
    'lev_highest_tropospheric_freezing_level': 'off',
    'lev_PV%3D-2e-06_(Km%5E2%2Fkg%2Fs)_surface': 'off',
    'lev_PV%3D2e-06_(Km%5E2%2Fkg%2Fs)_surface': 'off',
}
